/*

  resolve.c - Resolve and reflect database pointers

*/

#include <stdio.h>
#include "wnconsts.h"
#include "consts.h"
#include "typedef.h"
#include "extern.h"

static char *Id = "$Id: resolve.c,v 1.14 1997/08/05 14:23:46 wn Exp $";

static void DoPointers(Synset), DoFanPointers(Synset);
static void ResolveIt(Pointer, Synonym, Synset, int);
static void CheckSimilars(Synset), AddPointer(Synset, Pointer);
static Pointer ReflectPointer(Synset, Synset, unsigned char, short);
static Pointer ReflectSimPointer(Synset, Synset, short, int);
static Synonym FindUniqueSynonym(Pointer, Synset, int, int);
static Synonym FindUniqueSatellite(Pointer, Synset);
static int HasSimilar(Synset);

void ResolvePointers()
{
    Synset ss;

    /* Loop through all the synsets, resolving the pointers for
       each one. */

    for (ss = headss; ss; ss = ss->ssnext) {
	DoPointers(ss);
	if (ss->fans)
	    DoFanPointers(ss);
    }
}

static void DoPointers(Synset ss)
{
    Pointer p;
    Synonym syn;

    for (p = ss->ptrs; p; p = p->pnext) {

	if (p->status == UNRESOLVED) { /* don't do if already reflected */

	    /* Find the synset containing the word that this pointer
	       points to.  If there is no synset containing the word,
	       or if there is more than one synset which contains the
	       word, print an error message.  If the synset is the head
	       of a cluster, look for target synset in cluster's fan
	       synsets. */

	    if (ss->clusnum &&
		(p->ptype != PERTPTR && p->ptype != ATTRIBUTE)) {
		if
		    /* Resolve antonym in this cluster */

		    ((p->phead && p->ptype == ANTPTR &&
		      (syn = FindUniqueSynonym(p, ss, FALSE, ss->clusnum))) ||

		     /* Resolve pointer to another cluster (negative 
			cluster number to indicate another cluster) */

		     (p->phead &&
		      (syn = FindUniqueSynonym(p, ss, FALSE, -ss->clusnum))) ||

		     /* Resolve to fan synset of this cluster */

		     (!p->phead &&
		      (syn = FindUniqueSynonym(p, ss, TRUE, ss->clusnum))))
			
			ResolveIt(p, syn, ss, FALSE);

	    } else if (syn = FindUniqueSynonym(p, ss, FALSE, 0))
		ResolveIt(p, syn, ss, FALSE);
		    
	    /* If here, pointer is unresolved - no synset contains word
	       with same sense and file number. */
	    
	    if (p->status == UNRESOLVED) {
		fprintf(logfile, "Unresolved pointer: ");
		if (ss->filenum != p->pfilenum)
		    fprintf(logfile, "%s:", filelist[p->pfilenum].fname);
		fprintf(logfile,
			"%s in file %s, line %d\n",
			PrintPointer(p),
			filelist[ss->filenum].fname,
			ss->lineno);
		errcount++;
	    }
	}
    }
}

static void DoFanPointers(Synset clusheadss)
{
    Pointer p;
    Synonym syn;
    Synset fan;

    for (fan = clusheadss->fans; fan; fan = fan->ssnext) {

	for (p = fan->ptrs; p; p = p->pnext) {

	    if (p->status == UNRESOLVED) { /* don't do if already reflected */

		/* Fan synsets may contain pointers to other head
		   synsets.  Resolve fan synset pointers to other
		   clusters and reflect the cluster pointed to to this
		   fan's head synset (pass ResolveIt 'fan' rather than
		   'clusheadss'). */
		
		/* Find the synset containing the word that this pointer
		   points to.  If there is no synset containing the word,
		   or if there is more than one synset which contains the
		   word, print an error message. */

		if
		    /* Resolve to another cluster */
		    
		    ((p->phead &&
		      (syn = FindUniqueSynonym(p,
					       fan,
					       FALSE,
					       -clusheadss->clusnum))) ||

		     /* Resolve to this cluster */

		     (!p->phead &&
		      (syn = FindUniqueSynonym(p,
					       fan,
					       FALSE,
					       clusheadss->clusnum))) ||

		     /* Resolve to another fan in this cluster */

		     (!p->phead && p->ptype == ANTPTR &&
		      (syn = FindUniqueSynonym(p,
					       fan,
					       TRUE,
					       clusheadss->clusnum))))
#ifdef FOOP
			ResolveIt(p, syn, clusheadss, TRUE);
#endif
			ResolveIt(p, syn, fan, TRUE);
		
		/* If here, pointer is unresolved - no synset contains word
		   with same sense and file number. */
		
		if (p->status == UNRESOLVED) {
		    fprintf(logfile, 
			    "Unresolved fan pointer: %s in file %s, line %d\n",
			    PrintPointer(p),
			    filelist[clusheadss->filenum],
			    fan->lineno);
		    errcount++;
		}
	    }
	}
    }

    CheckSimilars(clusheadss);	/* make sure all satellites are SIMILARs */
}

static void CheckSimilars(Synset clusheadss)
{
    Synset fan;
    Pointer ref, lastp;

    for (fan = clusheadss->fans; fan; fan = fan->ssnext) {
	if (!HasSimilar(fan)) {	/* no similar in fan synset */
	    fprintf(logfile,
		    "Warning: Missing similar pointer in file %s, line %d\n",
		    filelist[clusheadss->filenum],
		    fan->lineno);

	    /* Automatically reflect pointer from fan to cluster head */

	    ref = CreatePointer(NULL,
				NULL,
				clusheadss->filenum,
				0,
				0,
				SIMPTR,
				0,
				0);
	    ref->psynset = clusheadss;
	    ref->status = RESOLVED;
	    AddPointer(fan, ref); /* add pointer to end of chain */

	    /* Now create pointer from cluster head to satellite */

	    ref = CreatePointer(NULL,
				NULL,
				clusheadss->filenum,
				0,
				0,
				SIMPTR,
				0,
				0);
	    ref->psynset = fan;
	    ref->status = RESOLVED;
	    AddPointer(clusheadss, ref); /* add pointer to end of chain */
	}
    }
}

static void ResolveIt(Pointer p, Synonym syn, Synset ss, int fanss)
{
    Pointer ref, lastp;
    
    p->psynset = syn->ss; /* point to target synset */
    
    p->towdnum = (p->fromwdnum != ALLWORDS) ? syn->sswdnum : ALLWORDS;

    /* If this pointer needs to be reflected, see if
       the lexicographer put in the reflection.  If not,
       automatically reflect the pointer. */

    if (ptrreflects[p->ptype]) {
	if (p->ptype == SIMPTR && ss->clusnum)
	    ref = ReflectSimPointer(syn->ss, ss, p->fromwdnum, fanss);
	else	
	    ref = ReflectPointer(syn->ss,
				 ss,
				 ptrkind(ptrreflects[p->ptype]),
				 p->fromwdnum);
	if (!ref) {
	    ref = CreatePointer(NULL,
				NULL,
				ss->filenum,
				0,
				0,
				ptrkind(ptrreflects[p->ptype]),
				p->towdnum,
				p->fromwdnum);
	    ref->psynset = ss;
	    ref->status = RESOLVED;
	    AddPointer(syn->ss, ref); /* add pointer to end of chain */
	}
    }
    p->status = RESOLVED;
}

static Pointer ReflectPointer(Synset ss, Synset tss,
			      unsigned char ptype, short wdnum)
{

    Pointer p;
    Synonym syn;

    for (p = ss->ptrs; p; p = p->pnext) {

	for (syn = tss->syns; syn; syn = syn->synnext) {

	    if (syn->word == p->pword &&
		syn->sensenum == p->psensenum &&
		tss->filenum == p->pfilenum &&
		p->ptype == ptype) {

		if (wdnum == ALLWORDS || syn->sswdnum == wdnum) {
		    p->psynset = tss;	/* point to target synset */
		    p->towdnum = wdnum;	/* set target word number */
		    p->status = RESOLVED;
		    return(p);
		}
	    }
	}
    }
    return(NULL);
}

static Pointer ReflectSimPointer(Synset ss, Synset tss,
				 short wdnum, int fanss)
{
    Pointer p;
    Synonym syn;

    /* 'fanss' is used to make sure we've found a pointer to the
       right synset.  In a cluster structure, head synsets have
       similar pointers to all fan synsets, and all fan synsets have
       a similar pointer to the cluster head.  These pointers are
       specified in lower case.  When ReflectSimPointer is called,
       fanss is FALSE when trying to match head/fan pointers within
       a single cluster.  It is TRUE when trying to match a pointer
       from a fan to a different cluster head. */

    syn = tss->syns;		/* look for pointer to head word */

    for (p = ss->ptrs; p; p = p->pnext) {
	if (syn->word == p->pword &&
	    p->phead == fanss &&
	    syn->sensenum == p->psensenum &&
	    tss->filenum == p->pfilenum &&
	    p->ptype == SIMPTR) {

	    if (wdnum == ALLWORDS || syn->sswdnum == wdnum) {
		p->psynset = tss; /* point to target synset */
		p->towdnum = wdnum; /* set target word number */
		p->status = RESOLVED;
		return(p);
	    }
	}
    }
    return(NULL);
}

static Synonym FindUniqueSynonym(Pointer p, Synset ss, int fanss, int clusnum)
{
    SynList sl;
    Synonym syn = NULL;
    int i = 0;
	    
    for (sl = p->pword->syns; sl; sl = sl->snext)
	if (sl->psyn->sensenum == p->psensenum &&
	    sl->psyn->ss->filenum == p->pfilenum &&
	    sl->psyn->ss->isfanss == fanss) {

	    if (p->pslite) {	/* found cluster head */

		/* Pointer to a satellite of this cluster head */

		if (syn = FindUniqueSatellite(p, sl->psyn->ss))
		    i++;

	    } else if
		
		/* Pointer to a headword (sswdnum < 0) in this cluster */

		(((clusnum > 0) && (sl->psyn->ss->clusnum == clusnum) &&
		  (sl->psyn->sswdnum < 0) && (fanss == FALSE)) ||

		 /* Pointer to a headword in another cluster */

		 ((clusnum < 0 ) && (sl->psyn->ss->clusnum != -clusnum) &&
		  (sl->psyn->sswdnum < 0) && (fanss == FALSE)) ||

		 /* Pointer to a fan synset in this part of cluster */

		 ((clusnum > 0 ) && (sl->psyn->ss->fans == ss) &&
		  (fanss == TRUE)) ||

		 /* Pointer to a fan synset in other part of cluster */

		 ((clusnum > 0 ) && (sl->psyn->ss->clusnum == clusnum) &&
		  (fanss == TRUE)) ||

		 /* Not a cluster, previous test suffice */

		 (clusnum == 0)) {
#ifdef FOOP

		    /* If one synonym has been found already, and
		       another has just been found, if they are in
		       the same synset it's ok.  (Synonym has been
		       entered in one synset with alternate or
		       non-lower-case spelling) If synset is not
		       the same, then the pointer is ambiguous. */

		    if (syn && (syn->ss == sl->psyn->ss))
			continue;
		    else {
#endif
			syn = sl->psyn;	/* get snyonym structure for word */
			i++;
#ifdef FOOP
		    }
#endif
		}
	}

    if (i > 1) {
	fprintf(logfile, "Ambiguous pointer: %s in file %s, line %d\n",
		PrintPointer(p),
		filelist[ss->filenum],
		ss->lineno);
	errcount++;
    }
    return(i == 1 ? syn : NULL);
}

static Synonym FindUniqueSatellite(Pointer p, Synset ss)
{
    Synset slite;
    Synonym syn;

    for (slite = ss->fans; slite; slite = slite->ssnext)
	for (syn = slite->syns; syn; syn = syn->synnext)
	    if ((p->pslite == syn->word) &&
		(p->pslite_sense == syn->sensenum))
		return(syn);
    return(NULL);
}

static void AddPointer(Synset ss, Pointer p)
{
    Pointer lastp;

    if (ss->ptrs) {
	for (lastp = ss->ptrs; lastp->pnext; lastp = lastp->pnext);
	lastp->pnext = p;
    } else
	ss->ptrs = p;
}

static int HasSimilar(Synset fan)
{
    Pointer p;

    for (p = fan->ptrs; p; p = p->pnext) {
	if (p->ptype == SIMPTR)
	    return(1);
    }
    return(0);
}

/*
  Revision log: (since version 1.5)

  $Log: resolve.c,v $
 * Revision 1.14  1997/08/05  14:23:46  wn
 * added function prototypes, other small cleanups
 *
 * Revision 1.13  1994/09/27  19:53:01  wn
 * changed to use filelist as structure
 *
 * Revision 1.12  1993/04/21  12:40:29  wn
 * added new ATTRIBUTE pointer (=)
 *
 * Revision 1.1  91/09/11  14:49:59  wn
 * Initial revision
 * 
*/
